/*
** avfilter_sig.c - written by vesely in milano on 22jan2004
** utility to send a signal to avfilter if it is running
*/

#if defined(HAVE_CONFIG_H)
#include <config.h>
#else
#error config.h missing
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/types.h>
#if TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#include <time.h>
#endif
#endif
#include <signal.h>
#if HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#if HAVE_FCNTL_H
#include <fcntl.h>
#endif

/* COURIER */
//#include "unconfig.h"
//#include <courier/courier.h> /* for config_localfilename */
//#if defined(HAVE_CONFIG_H)
//#include "unconfig.h"
//#include <config.h>
//#endif


#include "avfiledefs.h"

char const syntax_string1[] =
	"    avfilter_sig --help | --version";
char const syntax_string2[] =
	"    avfilter_sig [-p pid] [reload [config | virus]]";

// format string - expect args: syntax_strings 1 and 2,
// pidfile, pipefile, configfile
char const help_string[] =
	PACKAGE_STRING "\n"
	"\n"
	"syntax:\n"
	"   %s\n"
	" or\n"
	"   %s\n"
	"\n"
	"In the second form, avfilter_sig sends a signal to avfilter so that the\n"
	"latter reloads its configuration file, the virus data base, or both. The\n"
	"configuration file of avfilter can be overridden for testing in its own\n"
	"command line, otherwise it is configured as\n"
	"\t%s\n"
	"The process id of avfilter can be overridden for testing with -p, otherwise\n"
	"it is read from the pid file configured as\n"
	"\t%s\n"
	"A named pipe is used to read the response. Existence of the pipe is tested\n"
	"to avoid concurrent requests. When the -p option is used avfilter_sig uses\n"
	"\"%s\" as the pipe name (which works with tavfilter.)\n"
	"For regular operations, avfilter_sig uses the pipe name configured as\n"
	"\t%s\n"
	"\n"
	"The signal sent is\n"
	"HUP  -if neither conf nor virus is specified,\n"
	"USR1 -for conf, to reload configuration,\n"
	"USR2 -for virus, to reload the virus data base.\n"
	"\n"
	"Exit status is 0 if ok, 1 for daemon error, higher for other errors\n"
	"including daemon not running, pipe already in use, etc.\n";

static int signal_timed_out = 0,
	signal_break = 0;

static int keep_running(int timeout)
{
	if (signal_break > 0)
	{
		fprintf(stderr,
			"avilter_sig: caught %s signal\n", 
			signal_break == SIGPIPE ? "PIPE" :
			signal_break == SIGINT ? "INT" :
			signal_break == SIGQUIT ? "QUIT" :
			signal_break == SIGTERM ? "TERM" : "unexpected");
		signal_break = -1;
	}

	if (signal_timed_out > 0)
	{
		if (timeout > 0)
		{
			fprintf(stderr,
				"avfilter_sig: terminated after timeout [%d seconds]\n",
				timeout);
			signal_timed_out = -1;
		}
		else
			signal_timed_out = timeout;
	}
	
	return signal_timed_out == 0 && signal_break == 0;
}

static void sig_catcher(int sig)
{
	switch(sig)
	{
		case SIGALRM:
			signal_timed_out = 1;
			break;
		
		case SIGPIPE:
		case SIGINT:
		case SIGQUIT:
		case SIGTERM:
			signal_break = sig;
			break;
		default:
			break;
	}
}

static void init(void)
{
	struct sigaction act;
	memset(&act, 0, sizeof act);
	sigemptyset(&act.sa_mask);
		
	act.sa_flags = 0;
	act.sa_handler = sig_catcher;
	sigaction(SIGALRM, &act, NULL);
	sigaction(SIGPIPE, &act, NULL);
	sigaction(SIGINT, &act, NULL);
	sigaction(SIGTERM, &act, NULL);
}

static int kill_file(char const *pipe_fname)
{
	int rtc = 0;
	if (unlink(pipe_fname))
	{
		fprintf(stderr, "avfilter_sig: cannot remove %s: %s\n",
			pipe_fname, strerror(errno));
		rtc = 2;
	}
	return rtc;
}

static int verify_pid(pid_t pid, int pid_is_arg, char const *pid_from)
{
	errno = 0;
	if (kill(pid, 0))
	{
		int kill_pid_file = errno != EPERM;
		fprintf(stderr,
			"avfilter_sig: cannot verify process %d (read from %s): %s\n",
			(int)pid,
			pid_from,
			strerror(errno));
		if (pid_is_arg)
			fprintf(stderr, "(would%s have killed the pid file)\n",
				kill_pid_file ? "" : " not");
		else if (kill_pid_file)
			unlink(AVFILTER_PID_FILE);
		return 2;
	}
	return 0;
}

int main(int argc, char *argv[])
{
	int i;
	int rtc = 0, rrtc;
	int sig = 0;
	FILE *fp;
	char buf[1024];
	pid_t pid;
	int fifo_already_existed = 0;
	int fd = -1;
	char const* pipe_fname = AVFILTER_PIPE_FILE;
	char const* pid_option = NULL;
	char const* pid_from = AVFILTER_PID_FILE;
	mode_t perms;
	uid_t const my_euid = geteuid();
	
	init();
	
	for (i = 1; i < argc; ++i)
	{
		char *const arg = argv[i];
		if (strcmp(arg, "--help") == 0)
		{
			char *fname = COURIER_SYSCONF_INSTALL "/filters/avfilter.conf";
			if (fname == NULL)
			{
				fputs("avfilter_sig: config_localfilename returned NULL",
					stderr);
				return 2;
			}
			fprintf(stdout,
				help_string,
				syntax_string1,
				syntax_string2,
				fname,
				AVFILTER_PID_FILE,
				AVFILTER_PIPE_TEST,
				AVFILTER_PIPE_FILE);
			return 0;
		}

		if (strcmp(arg, "--version") == 0)
		{
			fprintf(stdout, "%s\n", PACKAGE_STRING);
			return 0;
		}
		
		if (strcmp(arg, "reload") == 0)
			(void)0;
		else if (strcmp(arg, "config") == 0)
			sig = sig ? SIGHUP : SIGUSR1;
		else if (strcmp(arg, "virus") == 0)
			sig = sig ? SIGHUP : SIGUSR2;
		else if (strcmp(arg, "-p") == 0)
		{
			if (++i >= argc)
			{
				fputs("avfilter_sig: option -p requires an argument\n", stderr);
				return 2;
			}
			pid_option = argv[i];
			pipe_fname = AVFILTER_PIPE_TEST;
			pid_from = "command line option \"-p\"";
		}
		else
		{
			fprintf(stderr, "%s: invalid: %s\nSyntax:\n%s\nor\n%s\n",
				argv[0], arg, syntax_string1, syntax_string2);
			return 2;
		}
	}
		
	if (pid_option)
		fp = NULL;
	else if ((fp = fopen(AVFILTER_PID_FILE, "r")) == NULL)
	{
		if (errno == ENOENT)
			fprintf(stderr,
				"avfilter_sig: avfilter not running: %s not found\n",
				AVFILTER_PID_FILE);
		else
			fprintf(stderr, "avfilter_sig: cannot read %s: %s\n",
				AVFILTER_PID_FILE, strerror(errno));
		return 2;
	}
	
	errno = 0;
	if (pid_option)
	{
		i = strlen(pid_option);
		if ((unsigned)i > sizeof buf)
			i = sizeof buf;
		memcpy(buf, pid_option, i);
	}
	else
		i = fread(buf, 1, sizeof buf, fp);
	if (i > 0)
	{
		unsigned s = i;
		char *t;
		unsigned long l;
		if (s >= sizeof buf)
			s = sizeof buf - 1;
		buf[s] = 0;
		l = strtoul(buf, &t, 0);
		if (t == NULL || t == &buf[0] ||
			(*t != 0 && *t != '\n') || l <= 0)
		{
			fprintf(stderr, "avfilter_sig: invalid pid \"%s\" from %s\n",
				buf, pid_from);
			i = 0;
		}
		
		pid = (pid_t)l;
	}
	else
	{
		fprintf(stderr, "avfilter_sig: cannot read in %s: %s\n",
			pid_from, strerror(errno));
	}
	if (fp)
		fclose(fp);
	if (i <= 0)
		return 2;

	if (verify_pid(pid, pid_option != NULL, pid_from))
		return 2;
	
	/* set perms for mkfifo */
	perms = S_IRUSR | S_IWUSR;
	if (my_euid != MAILUID && my_euid != 0)
		perms |= S_IWGRP;	
	
	if (mkfifo(pipe_fname, perms))
	{
		rtc = 2;
		if (errno == EEXIST)
		{
			struct stat st;

			fifo_already_existed = 1;
			if (stat(pipe_fname, &st) == 0)
			{
				unsigned long diff = time(NULL) - st.st_mtime;
				mode_t const mode = st.st_mode & S_IFMT;
				
				/* if it's stale we can reuse it... */
				if (diff > 8UL * 3600UL)
				{
					rtc = 0;
				}
				else if (mode == S_IFIFO)
				{
					char *msg = &buf[0];
					if (diff > 3600UL)
					{
						msg += sprintf(msg, "%ldh ", diff/3600UL);
						diff %= 3600UL;
					}
					if (diff > 60UL)
					{
						msg += sprintf(msg, "%ldm ", diff/60);
						diff %= 60UL;
					}
					sprintf(msg, "%lds", diff);
					
					fprintf(stderr, "avfilter_sig: %s already in use [%s old]\n",
						pipe_fname, buf);
				}
				
				/* ...can reuse fifo? */
				if (mode != S_IFIFO)
				{
					rtc = 2;
					fprintf(stderr, "avfilter_sig: %s already in use as %s\n",
						pipe_fname,
						mode == S_IFCHR ? "character special" :
						mode == S_IFDIR ? "directory" :
						mode == S_IFBLK ? "block special" :
						mode == S_IFREG ? "regular file" :
							"other file type");
				}
				else if (rtc == 0) /* try reusing stale pipe */
				{
					if ((st.st_mode & (S_IRWXU | S_IRWXG | S_IRWXO)) != perms)
						chmod(pipe_fname, perms);
				}
			}
			else
			{
				fprintf(stderr,
					"avfilter_sig:\n"
					"this shouldn't happen: cannot create pipe because already\n"
					"exists %s\n"
					"however stat reports it's not found... race condition?\n"
					"Please report this bug to " PACKAGE_BUGREPORT "\n",
						pipe_fname);
			}
		}
		else
		{
			fprintf(stderr, "avfilter_sig: cannot mkfifo %s: %s\n",
				pipe_fname, strerror(errno));
		}
		
		if (rtc)
			return rtc;
	}
	
	/*
	** if not running with courier ids, change the group owner or permissions
	** to make sure avfilter can write to the pipe
	*/
	if (my_euid != MAILUID &&
		chown(pipe_fname, my_euid ? -1 : MAILUID, MAILGID))
	{
		int const save_errno = errno;
		if (chmod(pipe_fname, perms |= S_IWOTH))
			fprintf(stderr,
				"avfilter_sig: failed to change group ownership of pipe\n"
				"%s (reason: %s)\n"
				"cannot even make it world writable: %s\n",
					pipe_fname, strerror(save_errno), strerror(errno));
	}
	

	/*
	** file must be opened for reading before we send a signal
	** to avfilter, otherwise it may get an ENXIO error when
	** trying to open it - for that we'll check if pipe EOF
	** occurs before avfilter had a chance to open or after
	** having read any data.
	*/
	rrtc = 0;
	errno = 0;
	fd = open(pipe_fname, O_RDONLY
#if defined(O_NONBLOCK)
		| O_NONBLOCK
#elif defined(O_NDELAY)
		| O_NDELAY
#else
#error neither O_NONBLOCK nor O_NDELAY defined
#endif
	);
	if (fd < 0)
	{
		fprintf(stderr, "avfilter_sig: cannot open %s for reading: %s\n",
			pipe_fname, strerror(errno));
		rrtc = 2;
	}
	
	/* Take out of NDELAY mode */
	if (rrtc == 0)
	{
		errno = 0;
		if (fcntl(fd, F_SETFL, 0) == -1)
		{
			fprintf(stderr,
				"avfilter_sig: cannot reset file descriptor flags: %s\n",
				strerror(errno));
			close(fd);
			rrtc = 2;
		}
	}

	if (rrtc == 0)
	{
		if (sig == 0)
			sig = SIGHUP;
		errno = 0;
		if (keep_running(0))
		{
			if (kill(pid, sig))
			{
				fprintf(stderr,
					"avfilter_sig: cannot signal to process %d (read from %s): %s\n",
					(int)pid,
					pid_from,
					strerror(errno));
				close(fd);
				rrtc = 2;
			}
		}
		else
			rrtc = 2;
	}

	if (rrtc == 0)
	{
		static const int timeout = 900;
		char *const ebuf = &buf[sizeof buf - 1];
		char *sbuf = &buf[0];
		int read_any = 0, read_complete = 0;
		
		alarm(timeout);
		while (keep_running(timeout))
		/*
		** read until timeout, error or empty line
		*/
		{
			int p;
			
			errno = 0;
			p = read(fd, sbuf, 1);
			if (p < 0)
			{
				switch (errno)
				{
					case EAGAIN:
					case EINTR:
						if (verify_pid(pid, pid_option != NULL, pid_from) == 0)
							continue;
						/* else thru */
					default:
						break;
				}
				fprintf(stderr, "avfilter_sig: cannot read %s: %s\n",
					pipe_fname,
					strerror(errno));
				rrtc = 2;
				break;
			}
			else if (p == 0) /* pipe EOF */
			{
				if (!read_any)
				{
					int const residual_timeout = alarm(1) - 1;
					pause();
					if (keep_running(residual_timeout > 0 ? 0 : timeout) &&
						verify_pid(pid, pid_option != NULL, pid_from) == 0)
					{
						alarm(residual_timeout);
						continue;
					}
				}
				else
					fprintf(stderr,
						"avfilter_sig: unexpected EOF on %s\n",
						pipe_fname);
				rrtc = 2;
				break;
			}
			
			read_any = 1;
			p = *sbuf++;
			if (p == '\n' || sbuf >= ebuf)
			{
				char *bbuf = &buf[0];
				unsigned sz = sbuf - bbuf, szl;
				*sbuf = 0;
				if (sz >= (szl = 4U) && memcmp(bbuf, "ERR:", szl) == 0 ||
					sz >= (szl = 6U) && memcmp(bbuf, "ALERT:", szl) == 0)
				{
					bbuf += szl;
					while (bbuf < sbuf && *bbuf == ' ')
						++bbuf;
					rtc = 1;
				}                            //     1234567890123
				else if (sz >= 13U && memcmp(bbuf, "END_OF_REPORT", 13U) == 0)
				{
					read_complete = 1;
					break;
				}
				fputs(bbuf, stdout);
				sbuf = &buf[0];
			}
		}
		
		if (!read_any)
		{
			fputs("avfilter_sig: no data read from pipe\n", stderr);
			rrtc = 2;
		}
		else if (!read_complete)
		{
			fputs("avfilter_sig: read was not complete\n", stderr);
			rrtc = 2;
		}
		
		if (close(fd))
		{
			fprintf(stderr, "avfilter_sig: cannot close %s: %s\n",
				pipe_fname, strerror(errno));
			rrtc = 2;
		}

		alarm(0);
		if (!keep_running(timeout))
			rrtc = 2;
	}	

	if (kill_file(pipe_fname))
		rrtc = 2;
	
	if (rtc == 0)
		rtc = rrtc;
	
	return rtc;
}
