/*
* run_get_ides.c - written by vesely in milan on 27 feb 2008
*
* run the script get_ides.sh
* install:
*   gcc -W -O -DNDEBUG -o /usr/local/sbin/run_get_ides run_get_ides.c
*   chmod ug+s /usr/local/sbin/run_get_ides
* debug: (lots of warnings)
*   gcc '-DSCRIPTFILE="./test_rgi.sh"' -DTESTSMALL_BUFFER \
*      -W -Wall -O0 -g -o trgi run_get_ides.c
*/
#if !defined SCRIPTFILE
#define SCRIPTFILE "/usr/local/sbin/get_ides.sh"
#endif

static char const usage[] =
"usage:\n"
"\trun_get_ides -h\n"
"\trun_get_ides [-D uid/gid] [-t timeout] [-s skip-code] [--] ...\n"
"The first format is only useful to print this note.\n"
"\n"
"The second format is used for invoking " SCRIPTFILE ".\n"
"This executable can be installed suid/sgid, and optionally takes the -D switch\n"
"for specifying target uid/gid for managing the antivirus files. Logging and\n"
"changing the subject on failure are the added benefits.\n"
"Skip-code is numeric, typically 304.  If the script's subject has such code\n"
"then nothing is output.  If no skip option is specified, xfilter is assumed.\n"
"All unrecognized arguments are passed to the script, -- may be used to mark end\n"
"of arguments.\n";

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <syslog.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <assert.h>

static volatile int
	signal_child = 0,
	signal_timed_out = 0,
	signal_break = 0;

#if !defined(NDEBUG)
static const char *get_signal_name(int sig)
{
	switch(sig)
	{
		case SIGALRM: return "SIGALRM";
		case SIGHUP:  return "SIGHUP";
		case SIGPIPE: return "SIGPIPE";
		case SIGINT:  return "SIGINT";
		case SIGQUIT: return "SIGQUIT";
		case SIGTERM: return "SIGTERM";
		case SIGCHLD: return "SIGCHLD";
		default:      return "unknown";
	}
}
#endif

static void sig_catcher(int sig)
{
#if !defined(NDEBUG)
	char buf[80];
	unsigned s = snprintf(buf, sizeof buf,
		"run_get_ides[%d]: received signal %d (%s)\n",
		(int)getpid(), sig, get_signal_name(sig));
	if (s >= sizeof buf)
	{
		buf[sizeof buf - 1] = '\n';
		s = sizeof buf;
	}
	write(2, buf, s);
#endif
	switch(sig)
	{
		case SIGALRM:
			signal_timed_out = 1;
			break;

		case SIGHUP:
		case SIGPIPE:
		case SIGINT:
		case SIGQUIT:
		case SIGTERM:
			signal_break = 1;
			break;

		case SIGCHLD:
			signal_child = 1;
			break;

		default:
			break;
	}
}

#if 0
static void reset_signal(void)
{
	struct sigaction act;
	memset(&act, 0, sizeof act);
	sigemptyset(&act.sa_mask);
	act.sa_handler = SIG_DFL;

	sigaction(SIGALRM, &act, NULL);
	sigaction(SIGPIPE, &act, NULL);
	sigaction(SIGINT, &act, NULL);
	sigaction(SIGTERM, &act, NULL);
	sigaction(SIGHUP, &act, NULL);
	sigaction(SIGCHLD, &act, NULL);
}
#endif

static void set_signal(void)
{
	struct sigaction act;
	memset(&act, 0, sizeof act);
	sigemptyset(&act.sa_mask);
		
	act.sa_handler = sig_catcher;
	sigaction(SIGALRM, &act, NULL);
	sigaction(SIGPIPE, &act, NULL);
	sigaction(SIGINT, &act, NULL);
	sigaction(SIGTERM, &act, NULL);
	sigaction(SIGHUP, &act, NULL);
	sigaction(SIGCHLD, &act, NULL);
}

static void copy_from_to(FILE *from, FILE *to, char const *descr)
{
	unsigned char buf[4096];
	size_t r, w;
	int err = 0;
	
	while ((r = fread(buf, 1, sizeof buf, from)) > 0)
		if ((w = fwrite(buf, 1, r, to)) != r)
		{
			err = 1;
			break;
		}
	
	if (err || ferror(from))
		syslog(LOG_CRIT, "Cannot copy %s: %s\n", descr, strerror(errno));
}

static size_t read_line(unsigned char **line)
{
	size_t bufsize = 1024;
	unsigned char *buf, *p, *ebuf;
	int ch;

	free(*line);
	*line = NULL;
	buf = p = (unsigned char*)malloc(bufsize);
	ebuf = buf + bufsize;

	if (buf)
		while ((ch = getchar()) != EOF)
		{
			if (p >= ebuf)
			{
				buf = (unsigned char*) realloc(buf, bufsize + bufsize);
				if (buf == NULL)
					break;
				
				p = buf + bufsize;
				bufsize += bufsize;
				ebuf = buf + bufsize;
			}
			
			if (ch == '\n')
			{
				if (p == buf || (ch = getchar()) == EOF)
					break; // empty line
				
				if (ch == '\n' || !isspace(ch))
				{
					ungetc(ch, stdin);
					break; // end of header
				}
				
				*p++ = '\n';
			}
			*p++ = (unsigned char)ch;
		}

	if (buf)
	{
		*p = 0;
		*line = buf;
		while (p > buf && isspace(p[-1]))
			*--p = 0;
	}
	
	return p > buf? p - buf: 0;
}

static const char subject[] = "Subject";

static int xfilter(char const *desc_rtc, FILE *fpout, int do_read)
{
	int found_subj = 0;
	size_t len;
	unsigned char *line = NULL;
	
	if (!desc_rtc) desc_rtc = "INVALID";
	
	if (do_read)
	{
		while ((len = read_line(&line)) > 0)
		{
			int here_subj = 0;
		
			if (!found_subj && len >= sizeof subject &&
				strncasecmp(subject, line, sizeof subject - 1) == 0)
			{
				unsigned char *p = line + sizeof subject - 1, *ep = line + len;

				while (p <= ep && isspace(*p))
					++p;
				if (*p == ':')
				{
					++p;
					while (p <= ep && isspace(*p))
						++p;
					fprintf(stdout, "%s: [%s] %s\n", subject, desc_rtc, p);
					here_subj = 1;
				}	
			}

			found_subj |= here_subj;
			if (!here_subj)
				fprintf(stdout, "%s\n", line);
		}
	
		if (!found_subj)
			fprintf(stdout, "%s: [%s]\n", subject, desc_rtc);

		fputc('\n', stdout); // end of headers

		copy_from_to(stdin, stdout, "body of message");
	
		fputs("\nLocal processing\n", stdout);
	}
	else
		fprintf(stdout, "\n\n%s\n", desc_rtc);

	if (fpout)
	{
		if (fseek(fpout, 0, SEEK_SET))
			syslog(LOG_CRIT, "Cannot rewind tmpfile: %s\n", strerror(errno));
		
		copy_from_to(fpout, stdout, "tmpfile");
		fputc('\n', stdout);
	}

	return !found_subj;
}

static void addtoenv(char const *name, int num)
{
	char value[64];
	sprintf(value, "=%d", num);
	char *freeonexit = (char*)malloc(strlen(name) + strlen(value) + 1);
	if (freeonexit)
		putenv(strcat(strcpy(freeonexit, name), value));
}



static char* run_script(int timeout, FILE *fpout, char *argv[])
{
	char* desc_rtc = NULL;
	int err[2];
		
	addtoenv("GET_IDES_TO", timeout);
	
	if (pipe(err) < 0)
		syslog(LOG_CRIT, "Cannot open pipe: %s\n", strerror(errno));
	else
	{
		pid_t const pid = fork();
		if (pid < 0)
			syslog(LOG_CRIT, "Cannot fork: %s\n", strerror(errno));
		else if (pid)
		{
#if TESTSMALL_BUFFER
			char buf[30];
#else
			char buf[2048];
#endif
			char *next = &buf[0];
			char *const first = &buf[0], *const last = &buf[sizeof buf - 2];
			
			close(err[1]);
			alarm(timeout + timeout);
			last[1] = 0; // terminator on forced newline
			while (signal_timed_out == 0 &&
				signal_break == 0 &&
				signal_child == 0)
			{
				int rd = read(err[0], next, last - next);
#if !defined NDEBUG
	printf("rd=%2d, next=%2ld\n", rd, next - first);
#endif
				if (rd > 0)
				{
					char *p = first, *br;
					next += rd;
				
					*next = next == last? '\n': 0; // force newline if full
				
					while ((br = strchr(p, '\n')) != NULL)
					{
						int level = LOG_CRIT;
						*br = 0;
				
						/*
						* e.g., "#An error\n##and a warning"
						*/
						while (*p == '#' && level < LOG_DEBUG)
							++p, ++level;
					
						if (*p)
						{
							char *ep = br;
							
							// trim
							while (ep-- > p)
								if (isspace(*(unsigned char*)ep))
									*ep = 0;
								else
									break;
							if (*ep)
								++ep;
							while (p < ep && isspace(*(unsigned char*)p))
								++p;
							
#if !defined NDEBUG
	printf("level=%d, len=%ld: %s\n", level, ep - p, p);
#endif
							if (p < ep)
							{
								syslog(level, "script: %s\n", p);

								// use last line as subject
								if ((size_t)(ep - p) > sizeof subject &&
									strncasecmp(subject, p, sizeof subject - 1) == 0)
								{
									p += sizeof subject - 1;
									while (p < ep && isspace(*(unsigned char*)p))
										++p;
									if (*p == ':')
									{
										while (++p < ep && isspace(*(unsigned char*)p))
											continue;
										if (p < ep)
										{
											free(desc_rtc);
											desc_rtc = strdup(p);
										}
									}
								}
								else
								{
									fwrite(p, 1, ep - p, fpout);
									putc('\n', fpout);
								}
								
							}
						}
						p = br + (br < last);  // +1 if not forced newline
						assert(first <= p && p <= next);
					}
				
					memmove(first, p, next - p);
					next -= p - first;
					assert(first <= next && next < last);
				}
				else if (rd == 0 || errno != EINTR && errno != EAGAIN)
				{
					if (rd)
						syslog(LOG_CRIT, "Pipe broken: %s\n", strerror(errno));
					break;
				}
			}
			alarm(0);
			if (signal_timed_out || signal_break)
			{
				kill(pid, SIGTERM);
			}
			close(err[0]);
			
			for (;;)
			{
				int status;
				pid_t wpid = wait(&status);
				if (wpid < 0 && errno != EAGAIN && errno != EINTR)
				{
					if (desc_rtc == NULL)
						desc_rtc = "GONE";
					syslog(LOG_CRIT,
						"Cannot wait " SCRIPTFILE "[%u]: %s (%s)\n",
						(unsigned)wpid, strerror(errno), desc_rtc);
					break;
				}
				else if (wpid == pid)
				{
					if (WIFEXITED(status))
					{
						int level, s_rtc = WEXITSTATUS(status);
						switch (s_rtc)
						{
							case 0: level = LOG_DEBUG; break;
							default: level = LOG_CRIT; break;
						}
						
						/*
						* return value
						*/
						if (desc_rtc == NULL)
							desc_rtc = s_rtc? "FAILED": "OK";
						syslog(level,
							SCRIPTFILE " exit status: %d (%s)\n", s_rtc, desc_rtc);
					}
					else if (WIFSIGNALED(status))
					{
						if (desc_rtc == NULL)
							desc_rtc = "ABORT";
						syslog(LOG_CRIT,
							SCRIPTFILE " terminated with signal %d, rtc=%s\n",
							WTERMSIG(status), desc_rtc);
					}
					else continue; // stopped?
					
					break;
				}
			}
		}
		else // child process
		{
			close(0);
			open("/dev/null", O_RDONLY);
			close(1);
			dup(fileno(fpout));
			close(2);
			dup(err[1]);
			close(err[0]);
			close(err[1]);
			closelog();
			fclose(fpout);
			execv(SCRIPTFILE, argv);
			syslog(LOG_MAIL|LOG_CRIT, "run_get_ides: cannot execv: %s\n",
				strerror(errno));
			exit(33);
		}
	}
	return desc_rtc;
}

int main(int argc, char *argv[])
{
	char* desc_rtc = "ERROR";
	int i, uid = 0, gid = 0, uidgid = 0, timeout = 0, skip = 0,
		err = 0, pass_all = 0;
	FILE *fpout = NULL;
	static char const argerror[] = "invoked with wrong argument: ";

	char *xargv[32];
	size_t xargc = 0;
	time_t start, end;

	time(&start);	
	openlog("run_get_ides", LOG_PID, LOG_MAIL);
	xargv[xargc++] = SCRIPTFILE;
	
	for (i = 1; i < argc; ++i)
	{
		char *a = argv[i];
		int pass_it = 1;
		
		if (*a == '-' && !pass_all)
		{
			pass_it = 0;
			switch (*++a)
			{
				case 'D':
				case 't':
				case 's':
					if (i + 1 >= argc)
					{
						syslog(LOG_CRIT, "%s-%c requires a value\n", argerror, *a);
						++err;
					}
					else
					{
						char *t = NULL;
						int *value = *a == 'D'? &uid: *a == 't'? &timeout : &skip;
						
						a = argv[++i];
						*value = strtoul(a, &t, 10);
						if (t && *t == '/' && value == &uid)
						{
							gid = strtoul(t + 1, &t, 10);
							uidgid = 1;
							if (t && *t) t = NULL;
						}
						else if (t && *t || value == &uid) t = NULL;
						if (t == NULL)
						{
							syslog(LOG_CRIT, "%s%s is %s\n", argerror,
								value == &uid? "uidgid": "timeout", a);
							++err;
						}
					}
					break;
				
				case '-':
					pass_all = 1;
					break;
					
				case 'h':
					puts(usage);
					return 0;

				default:
					pass_it = 1;
					break;
			}
		}
		
		if (pass_it)
			if (xargc + 1 < sizeof xargv/ sizeof xargv[0])
				xargv[xargc++] = a;
			else
			{
				syslog(LOG_CRIT, "%scan pass max %lu arguments (found %d)\n",
					argerror, sizeof xargv/ sizeof xargv[0], argc);
				++err;
				break;
			}
	}
	
	xargv[xargc] = NULL;

	if (/* geteuid() == 0 && */ uidgid)
	{
		if (setgid(gid) || setuid(uid))
		{
			syslog(LOG_CRIT, "%sunable to set -D %d/%d: %s\n",
				argerror, uid, gid, strerror(errno));
			++err;
		}
	}
	
	if (timeout <= 0)
		timeout = 900;

	set_signal();
	if (err == 0)
	{
		struct stat buf;
		uid_t const me = geteuid();
		gid_t const myg = getegid();
		
		if (stat(SCRIPTFILE, &buf) != 0)
		{
			syslog(LOG_CRIT, "Cannot stat " SCRIPTFILE ": %s\n",
				strerror(errno));
		}
		else if (!S_ISREG(buf.st_mode) || !(
			((S_IXUSR & buf.st_mode) && (me == 0 || me == buf.st_uid)) ||
			((S_IXGRP & buf.st_mode) && (me == 0 || myg == buf.st_gid)) ||
			(S_IXOTH & buf.st_mode)))
		{
			syslog(LOG_CRIT, SCRIPTFILE " not executable by %d/%d\n",
				(int)me, (int)myg);
		}
		else
		{
			if ((fpout = tmpfile()) == NULL)
				syslog(LOG_CRIT, "Cannot create tmpfile: %s\n",
					strerror(errno));
			else
				desc_rtc = run_script(timeout, fpout, xargv);
		}
	}
	else
	{
		desc_rtc = "WON'T WORK";
	}

	if (skip)
	{
		if (atoi(desc_rtc) != skip) // don't skip
		{
			skip = 0;
			xfilter(desc_rtc, fpout, 0);
		}
	}
	else if (!isatty(fileno(stdin)))
		xfilter(desc_rtc, fpout, 1);

	if (fpout)
		fclose(fpout);
	
	time(&end);
	end -= start;
	//if (end >= timeout - timeout/4)
	if (skip == 0)
		printf("\ntimeout was %d, process took %d\n", timeout, (int)end);
	return 0;
}

